/*****************************************************
Chip type               : ATmega8L
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/

#include <mega8.h>

#define U_bat_min ((255 * 7 )/18)
#define U_bat_st1 ((255 * 10 )/18)
#define U_bat_st2 ((255 * 14 )/18)
#define U_bat_st3 ((255 * 16 )/18)
#define I_on PORTB.4
#define I_init 25
#define charge_led PORTB.2
#define ready_led PORTB.3
#define OCR_min 50
#define OCR_max 1023
#define TL494_run PORTB.0
#define battery_discharge PORTB.5 //not used in this version

#define A PORTC.3
#define B PORTC.4                 
#define C PORTC.5
#define D PORTD.0
#define E PORTD.1
#define F PORTD.2
#define G PORTD.3
#define DP PORTD.4

#define A1 PORTC.0
#define A2 PORTC.1
#define A3 PORTC.2

#define key_I_up PINB.6
#define key_I_dn PINB.7
#define key_Ah PIND.5

volatile unsigned int minutes;
volatile unsigned int timer_1, timer_2, timer_3, timer_4; 

volatile unsigned char I_ctr, I_def, I_curr ;

volatile unsigned char U_bat;

volatile eeprom unsigned char I_def_ee;

volatile bit charge_ok, Ah_999;

volatile unsigned int Ah_60;

volatile char data [3];
volatile unsigned char counter;

volatile unsigned int volts, amps, Ah;
volatile unsigned char display_trigger;




void ports_init () 
{
 DDRB = 0b00111111; PORTB = 0b11000000;
 DDRC = 0b00111111;
 DDRD = 0b01011111; PORTD = 0b00100000;
}

void TCNT0_init ()
{ TCCR0 |= 1<<CS01 | 1<<CS01;  TIMSK |= 1<<TOIE0;  }

void TCNT2_init () 
{ TCCR2 |= 1<<CS22 | 1<<CS21 | 1<< WGM21; OCR2 = 250; TIMSK |= 1<<OCIE2; }    // 31250 Hz   ctc s/125

void TCNT1_init ()  // 10 bit PWM, 8MHz, inv pwm on OC1A
{   
    TCCR1A |= 1<<COM1A1 | 1<<COM1A0 | 1<< WGM10 | 1<< WGM11;
    TCCR1B |= 1<<WGM12 | 1<<CS10;
}

void ACMP_init ()
{
 ACSR |= 1<<ACBG | 1<<ACIE | 1<<ACIS1;
}

void ADC_init ()
{
 ADMUX |= 1<<REFS1 | 1<<REFS0 | 1<<ADLAR | 6;
 ADCSRA |= 1<<ADEN | 1<<ADIE | 1<<ADPS2 | 1<<ADPS0; // 250 kHz 
 ADCSRA |= (1<<ADSC);
}

//___________________________________________________________________________________
void convert_data (unsigned int x)
{
 unsigned int temp,res;
 temp=x;
 res=temp/1000;//Calculate 1000-s
 data[2]=res;
 temp=temp-res*1000;
 res=temp/100;//Calculate 100-s  
 data[1]=res;
 temp=temp-res*100;
 res=temp/10;//Calculate 10-s
 data[0]=res;
}
 
void clear_display ()
{
 A=1; B=1; C=1; D=1; E=1; F=1; G=1; DP=1; A1=0; A2=0; A3=0;
}
 
void write_code ()
{
switch (data[counter])
   {
      case 0:            (A=0),(B=0),(C=0),(D=0),(E=0),(F=0);           break;
      case 1:            (B=0),(C=0);                                                   break;
      case 2:            (A=0),(B=0),(D=0),(E=0),(G=0);                     break;
      case 3:            (A=0),(B=0),(C=0),(D=0),(G=0);                     break;
      case 4:            (B=0),(C=0),(F=0),(G=0);                               break;
      case 5:            (A=0),(C=0),(D=0),(F=0),(G=0);                     break;
      case 6:            (A=0),(C=0),(D=0),(E=0),(F=0),(G=0);           break;
      case 7:            (A=0),(B=0),(C=0);                                         break;
      case 8:            (A=0),(B=0),(C=0),(D=0),(E=0),(F=0),(G=0); break;
      case 9:            (A=0),(B=0),(C=0),(D=0),(F=0),(G=0);           break;
      default:          (G=0);                                                             break;
   }
} 
 
void enable_digit  ()
{
 switch (counter)   
    {
     case 0: (A1 = 1); break;
     case 1: (A2 = 1); (DP=0); if (display_trigger == 2) {DP = Ah_999;}; break;
     case 2: if (data[2]) {(A3 = 1);} else {(A3 = 0);}  break;  
    } 
}

void  calculate_volts ()
{ volts = ((unsigned long) U_bat * 1800)/255; }
 
void  calculate_amps ()
{
   if (timer_3) {amps = ((unsigned long)I_def * 1000)/255;}
   else {amps = ((unsigned long)I_ctr * 1000)/255;}
}

void calculate_Ah ()
{ 
if (Ah <= 9999) {Ah_999 = 0; Ah = (Ah_60 / 6) * 10;} 
else {Ah_999 = 1; Ah = Ah_60 / 6;} 
}

void switch_display ()
{
    switch (display_trigger)
    {
     case 0: calculate_volts (); convert_data (volts); break;
     case 1: calculate_amps (); convert_data (amps); break;
     case 2: calculate_Ah ();  convert_data (Ah); break;
     default: break;
    }
}
//___________________________________________________________________________________
interrupt [TIM0_OVF] void t0_ovf ()
{
    switch_display ();
    clear_display ();
    write_code ();
    enable_digit  (); 
    
    if (counter > 2) {counter = 0;}
    else {counter ++;}
}

interrupt [TIM2_COMP] void t2_ctc ()
{
  static unsigned int t_1m = 0;
  
  t_1m ++; 
  
  if (t_1m > 7500) 
  {
    t_1m = 0; if (minutes < 65000) { minutes ++; }
    Ah_60 +=  (((unsigned int) I_ctr*10) / 255);  
  }
   
  
  if (timer_1) { timer_1 -- ; }
  if (timer_2) { timer_2 -- ; } 
  if (timer_3) { timer_3 -- ; } 
  if (timer_4) { timer_4 -- ; } 
  

  
}

interrupt [ANA_COMP] void acmp_fe ()
{ timer_1 = 650;  I_on = 0; TL494_run = 0; OCR1A = OCR_min; charge_led = 1; ready_led = 1; }

interrupt [ADC_INT] void Adc()
{
static unsigned int tmp_U = 0;
static unsigned char i_time = 255;
static unsigned char t_char = 10;

 if ((t_char))
    {
    if  (i_time) {tmp_U += ADCH; i_time--;}
    else 
        {
        I_ctr = (tmp_U / 255); 
        tmp_U = 0; i_time = 255; 
        t_char --; 
        
        if ((I_ctr < I_curr)&&(timer_1 == 0)&&(OCR1A < OCR_max)) {OCR1A ++;} else if (OCR1A > OCR_min) {OCR1A -- ;}
        
        if (t_char == 0) { ADMUX.MUX0 = 1; }  // switch to ADC7  
        }
   ADCSRA |= (1<<ADSC);   
 }
 
 else
 {   
  if  (i_time) {tmp_U += ADCH; i_time--;}
  else 
        {
        U_bat = (tmp_U / 255); 
        tmp_U = 0; i_time = 255; 
        
        ADMUX.MUX0 = 0;     // switch to ADC6
        t_char = 10;
        }
   ADCSRA |= (1<<ADSC);   
}

}
  
volatile enum led_n {green, red} led ;

void blinking_led (char led, unsigned int t_led)
{
static bit a;

if (led)
{
 ready_led = 0;   
 if (timer_2 == 0) { timer_2 = t_led; charge_led = a; a = ~a;}
} 
else 
{
charge_led = 0;
 if (timer_2 == 0) { timer_2 = t_led; ready_led = a; a = ~a;}
}
}

void charging ()
{
 if ((U_bat < U_bat_min)||(U_bat > ((255 * 17 )/18))) { I_on = 0; ready_led = 0; led = red; blinking_led (led, 10); timer_1 = 625; OCR1A = OCR_min;}
 else if ((timer_1 == 0)&&(charge_ok == 0)) 
 {  I_on = 1; TL494_run = 1;
    if (U_bat < U_bat_st1) { I_curr = I_def >> 3; ready_led = 0; charge_led = 1; }
    else if (U_bat < U_bat_st2) { I_curr = I_def; led = red; blinking_led (led, 60); }  
    else if (U_bat < U_bat_st3) { I_curr = I_def >> 2; led = green; blinking_led (led, 60); }
    else if ((U_bat >= U_bat_st3)||(Ah_60  > 12000)) { I_curr = I_def >> 5; ready_led = 1; charge_led = 0; charge_ok = 1; }
 }
 if ((U_bat < U_bat_st1)&&(charge_ok)) {charge_ok = 0; Ah_60 = 0;}
 if (charge_ok) { I_curr = I_def >> 5; ready_led = 1; charge_led = 0; }
}

void scan_keys ()
{
static bit tmp0, tmp1, tmp2, t;
static unsigned char count0, count1, count2;

if (tmp0 != key_I_up) {tmp0 = key_I_up; count0 = 0;} 
else {count0++;}
if (tmp1 != key_I_dn) {tmp1 = key_I_dn; count1 = 0;}
else {count1++;}
if (tmp2 != key_Ah) {tmp2 = key_Ah; t = 0; count2 = 0;}
else {count2++;}


if (count0 > 250) {if ((tmp0==0)&&(I_def<250)) {I_def += 1; timer_3 = 125; display_trigger = 1;} count0 = 0;}
if (count1 > 250) {if ((tmp1==0)&&(I_def>5)) {I_def -= 1; timer_3 = 125; display_trigger = 1;} count1 = 0;}
if (count2 > 250) {if ((tmp2==0)&&(t == 0)) { t = 1; if (display_trigger < 2) {display_trigger ++;} else {display_trigger = 0;}  count2 = 0;}}
}

void I_def_mem ()
{
static bit t_bit;

if (timer_3) {timer_4 = 1250; t_bit = 1;}
if ((timer_4 == 0)&&(t_bit)) {t_bit = 0; if (I_def_ee != I_def) { I_def_ee = I_def;}} 

}
                                      


void main(void)
{
ports_init ();
TCNT2_init ();
TCNT1_init ();
TCNT0_init ();
ACMP_init ();
ADC_init ();

timer_1 = 250;

if ((I_def_ee == 0)||(I_def_ee == 0xFF)) { I_def_ee = I_init; }
I_def = I_def_ee;

#pragma optsize-
WDTCR=0x1E;
WDTCR=0x0E;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

#asm("sei")

while (1)
{
charging ();
scan_keys ();
I_def_mem (); 
#asm ("wdr")      

}
}

